(* Beispielkomponente TCalendar mit Ergnzungen

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   J. Rathlev, Jun. 2004
   mit Kalenderwoche, Mai 2010
   *)

unit JRCalendar;

interface

uses Classes, Controls, Messages, Windows, Forms, Graphics, StdCtrls,
  ComCtrls, Grids, SysUtils, DateUtils, DateUtilEx;

// Definiere CompPalPage (siehe Register)
{$Include UserComps.pas }

const
  clLightRed = $87D0FF;

type
  TDayOfWeek = 0..6;

  TJRDateTimePicker = class(TDateTimePicker)
  public
    procedure Click; override;
    end;

  TJRCalendar = class(TCustomDrawGrid)
  private
    FFirstOfMonth,
    FMarkDate,
    FDate: TDateTime;
    FMonthOffset: Integer;
    FOnChange: TNotifyEvent;
    FReadOnly: Boolean;
    FStartOfWeek: TDayOfWeek;
    FUpdating: Boolean;
    FShowCurrentDate,
    FUseCurrentDate: Boolean;
    FIsHoliday : TDateQuery;
    function GetCellDate(ACol, ARow: Integer): TDateTime;
    function GetCellValue(ACol, ARow: Integer): integer;
    function GetDateElement(Index: Integer): Integer;
    procedure SetCalendarDate(Value: TDateTime);
    procedure SetMarkDate(Value: TDateTime);
    procedure SetDateElement(Index: Integer; Value: Integer);
    procedure SetStartOfWeek(Value: TDayOfWeek);
    procedure SetUseCurrentDate(Value: Boolean);
    function StoreCalendarDate: Boolean;
  protected
    procedure Change; dynamic;
    procedure ChangeMonth(Delta: Integer);
    procedure Click; override;
    function DaysThisMonth: Integer; virtual;
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    function IsLeapYear(AYear: Integer): Boolean; virtual;
    function SelectCell(ACol, ARow: Longint): Boolean; override;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
  public
    constructor Create(AOwner: TComponent); override;
    property CalendarDate: TDateTime  read FDate write SetCalendarDate stored StoreCalendarDate;
    property MarkDate: TDateTime  read FMarkDate write SetMarkDate;
    property CellValue[ACol, ARow: Integer]: integer read GetCellValue;
    property CellDate[ACol, ARow: Integer]: TDateTime read GetCellDate;
    procedure NextMonth;
    procedure NextYear;
    procedure PrevMonth;
    procedure PrevYear;
    procedure UpdateCalendar; virtual;
    property IsHoliday : TDateQuery write FIsHoliday;
  published
    property Align;
    property Anchors;
    property BorderStyle;
    property Color;
    property Constraints;
    property Ctl3D;
    property Day: Integer index 3  read GetDateElement write SetDateElement stored False;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FixedColor;
    property Font;
    property GridLineWidth;
    property Month: Integer index 2  read GetDateElement write SetDateElement stored False;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly: Boolean read FReadOnly write FReadOnly default False;
    property ShowHint;
    property StartOfWeek: TDayOfWeek read FStartOfWeek write SetStartOfWeek;
    property TabOrder;
    property TabStop;
    property UseCurrentDate: Boolean read FUseCurrentDate write SetUseCurrentDate default True;
    property ShowCurrentDate: Boolean read FShowCurrentDate write FShowCurrentDate default True;
    property Visible;
    property Year: Integer index 1  read GetDateElement write SetDateElement stored False;
    property OnClick;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnSelectCell;
    property OnStartDock;
    property OnStartDrag;
  end;

procedure Register;

implementation

procedure TJRDateTimePicker.Click;
begin
  inherited Click;
  end;

{ ------------------------------------------------------------------- }
constructor TJRCalendar.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  { defaults }
  FUseCurrentDate := True;
  FShowCurrentDate:=true;
  FixedCols := 1;
  FixedRows := 1;
  GridLineWidth:=0;
  Font.Style:=[fsBold];
  ColCount := 8;
  RowCount := 7;
  FStartOfWeek:=1; //  ISO 8601 = Mo
  ScrollBars := ssNone;
  Options := Options - [goRangeSelect] + [goDrawFocusSelected];
  FDate := Date;
  FMarkDate:=0;
  FIsHoliday:=nil;
  UpdateCalendar;
end;

procedure TJRCalendar.Change;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TJRCalendar.Click;
begin
  inherited Click;
  if CellValue[Col, Row]<> 0 then Day := CellValue[Col, Row];
end;

function TJRCalendar.IsLeapYear(AYear: Integer): Boolean;
begin
  Result := (AYear mod 4 = 0) and ((AYear mod 100 <> 0) or (AYear mod 400 = 0));
end;

function TJRCalendar.DaysThisMonth: Integer;
begin
  Result := DaysInAMonth(Year, Month);
end;

procedure TJRCalendar.DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
var
  TheText: string;
  fc,bc : TColor;
  val   : integer;
begin
  if ACol = 0 then begin
    if ARow=0 then begin
      TheText := 'KW';
      bc:=FixedColor;
      end
    else begin
      TheText := IntToStr(CellValue[ACol, ARow]);
      bc:=Color;
      end;
    fc:=clGray;
    end
  else if ARow=0 then begin
    TheText := ShortDayNames[CellValue[ACol, ARow]];
    fc:=clBlack; bc:=FixedColor;
    end
  else begin
    val:=CellValue[ACol, ARow];
    TheText := IntToStr(val and $1F);
    bc:=Color;
    if val and $20<>0 then begin  // Feiertag
      fc:=clRed;
      if val and $80<>0 then bc:=clLightRed;
      end
    else if val and $40<>0 then fc:=clGray
    else if val and $80<>0 then begin  // heute
      bc:=clNavy; fc:=clYellow;
      end
    else if val and $100<>0 then begin  // Marke
      bc:=clRed; fc:=clYellow;
      end
    else fc:=clBlack;
    end;
  with ARect, Canvas do begin
    Brush.Color:=bc; Font.Color:=fc;
    TextRect(ARect, Left + (Right - Left - TextWidth(TheText)) div 2,
     Top + (Bottom - Top - TextHeight(TheText)) div 2, TheText);
    end;
  end;

function TJRCalendar.GetCellDate(ACol, ARow: Integer): TDateTime;
begin
  if ARow = 0 then  Result:=0
  else Result:=FFirstOfMonth+FMonthOffset+ACol+(ARow-1)*7-1;
  end;

function TJRCalendar.GetCellValue(ACol, ARow: Integer): integer;
var
  DayNum : Integer;
  b      : boolean;
begin
  if ACol=0 then begin
    if ARow=0 then Result:=-1
    else begin  // KW
      DayNum := FMonthOffset + (ARow-1) * 7;  // Montag
      Result:=WeekOfTheYear(FFirstOfMonth+DayNum-1);
      end;
    end
  else if ARow = 0 then  { day names at tops of columns }
    Result := (StartOfWeek + ACol-1) mod 7 + 1
  else begin
    DayNum := FMonthOffset + (ACol-1) + (ARow-1) * 7;
    if (DayNum < 1) then Result := DayOf(FFirstOfMonth+DayNum-1) or $40
    else if (DayNum > DaysThisMonth) then Result := DayOf(FFirstOfMonth+DayNum-1) or $40
    else begin
      Result := DayNum;
      if assigned(FIsHoliday) then b:=FIsHoliday(FFirstOfMonth+DayNum-1) else b:=false;;
      if ((StartOfWeek + ACol-1) mod 7=0) or b then Result:=Result or $20;
      if FShowCurrentDate and SameDate(FFirstOfMonth+DayNum-1,Date) then Result:=Result or $80;
      if SameDate(FFirstOfMonth+DayNum-1,FMarkDate) then Result:=Result or $100;
      end;
    end;
  end;

function TJRCalendar.SelectCell(ACol, ARow: Longint): Boolean;
begin
  if ((not FUpdating) and FReadOnly) or (CellValue[ACol, ARow] = 0) then
    Result := False
  else Result := inherited SelectCell(ACol, ARow);
end;

procedure TJRCalendar.SetCalendarDate(Value: TDateTime);
begin
  FDate := Value;
  UpdateCalendar;
  Change;
end;

procedure TJRCalendar.SetMarkDate(Value: TDateTime);
begin
  FMarkDate := Value;
  UpdateCalendar;
  end;

function TJRCalendar.StoreCalendarDate: Boolean;
begin
  Result := not FUseCurrentDate;
end;

function TJRCalendar.GetDateElement(Index: Integer): Integer;
var
  AYear, AMonth, ADay: Word;
begin
  DecodeDate(FDate, AYear, AMonth, ADay);
  case Index of
    1: Result := AYear;
    2: Result := AMonth;
    3: Result := ADay;
    else Result := -1;
  end;
end;

procedure TJRCalendar.SetDateElement(Index: Integer; Value: Integer);
var
  AYear, AMonth, ADay: Word;
begin
  if Value > 0 then
  begin
    DecodeDate(FDate, AYear, AMonth, ADay);
    case Index of
      1: if AYear <> Value then AYear := Value else Exit;
      2: if (Value <= 12) and (Value <> AMonth) then AMonth := Value else Exit;
      3: if (Value <= DaysThisMonth) and (Value <> ADay) then ADay := Value else Exit;
      else Exit;
    end;
    if ADay>DaysInAMonth(AYear,AMonth) then ADay:=1;
    FDate := EncodeDate(AYear, AMonth, ADay);
    FUseCurrentDate := False;
    UpdateCalendar;
    Change;
  end;
end;

procedure TJRCalendar.SetStartOfWeek(Value: TDayOfWeek);
begin
  if Value <> FStartOfWeek then
  begin
    FStartOfWeek := Value;
    UpdateCalendar;
  end;
end;

procedure TJRCalendar.SetUseCurrentDate(Value: Boolean);
begin
  if Value <> FUseCurrentDate then
  begin
    FUseCurrentDate := Value;
    if Value then
    begin
      FDate := Date; { use the current date, then }
      UpdateCalendar;
    end;
  end;
end;

{ Given a value of 1 or -1, moves to Next or Prev month accordingly }
procedure TJRCalendar.ChangeMonth(Delta: Integer);
var
  AYear, AMonth, ADay: Word;
  NewDate: TDateTime;
  CurDay: Integer;
begin
  DecodeDate(FDate, AYear, AMonth, ADay);
  CurDay := ADay;
  if Delta > 0 then ADay := DaysInAMonth(AYear, AMonth)
  else ADay := 1;
  NewDate := EncodeDate(AYear, AMonth, ADay);
  NewDate := NewDate + Delta;
  DecodeDate(NewDate, AYear, AMonth, ADay);
  if DaysinAMonth(AYear, AMonth) > CurDay then ADay := CurDay
  else ADay := DaysInAMonth(AYear, AMonth);
  CalendarDate := EncodeDate(AYear, AMonth, ADay);
end;

procedure TJRCalendar.PrevMonth;
begin
  ChangeMonth(-1);
end;

procedure TJRCalendar.NextMonth;
begin
  ChangeMonth(1);
end;

procedure TJRCalendar.NextYear;
begin
  if IsLeapYear(Year) and (Month = 2) and (Day = 29) then Day := 28;
  Year := Year + 1;
end;

procedure TJRCalendar.PrevYear;
begin
  if IsLeapYear(Year) and (Month = 2) and (Day = 29) then Day := 28;
  Year := Year - 1;
end;

procedure TJRCalendar.UpdateCalendar;
var
  AYear, AMonth, ADay: Word;
begin
  FUpdating := True;
  try
    DecodeDate(FDate, AYear, AMonth, ADay);
    FFirstOfMonth := EncodeDate(AYear, AMonth, 1);
    FMonthOffset := 2 - ((DayOfWeek(FFirstOfMonth) - StartOfWeek + 7) mod 7); { day of week for 1st of month }
    if FMonthOffset = 2 then FMonthOffset := -5;
//    MoveColRow((ADay - FMonthOffset) mod 7, (ADay - FMonthOffset) div 7 + 1,False, False);
    Invalidate;
  finally
    FUpdating := False;
  end;
end;

procedure TJRCalendar.WMSize(var Message: TWMSize);
var
  GridLines: Integer;
begin
  GridLines := (ColCount+1) * GridLineWidth;
  DefaultColWidth := (Message.Width - GridLines) div ColCount;
  DefaultRowHeight := (Message.Height - GridLines) div RowCount;
end;

{ ---------------------------------------------------------------- }
procedure Register;
begin
  RegisterComponents(CompPalPage, [TJRCalendar]);
  RegisterComponents(CompPalPage, [TJRDateTimePicker]);
  end;

end.
