(* Delphi Komponente
   Eingabe von Zahlen
   ==================
   - Edit: Integer (dezimal oder hex)
   - Edit: Integer mit Bereichsberprfung
   - Edit: Fliekomma
   - Edit: Fliekomma mit Bereichsberprfung
   - ComboBox: Fliekomma

    2007 J. Rathlev 24147 Klausdorf
      Web:  www.rathlev-home.de
      Mail: info(a)rathlev-home.de

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Sep. 1998
         1.1 - Okt. 2001
         1.2 - Jun. 2005  : verbesserte Bereichsberprfung
         1.3 - May  2006  : AddValue-Routines
         1.4 - Apr. 2007  : Darstellung als Bin. Oktal, Dezimal und Hex
   *)

unit NumberEd;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  CommCtrl, ComCtrls, Dialogs, Menus, Forms, StdCtrls, WinMath;

// Definiere CompPalPage (siehe Register)
{$Include UserComps.pas }

const
  // from CommCtrl.h (Vers. 5)
  UDM_SETPOS32 = WM_USER+113;
  UDM_GETPOS32 = WM_USER+114;

  NumAccel = 3;

type
  TAccelArray = array [0..NumAccel-1] of TUDAccel;

type
  TNumFloatFormat = (ffGeneral, ffExponent, ffFixed, ffNumber, ffCurrency, ffNormalized);
  TAngleFormat = (fmDezGrad,fmDezMin,fmSekunden);
  TNumMode = (nmDecimal,nmHex,nmOctal,nmBin);

const
  NumBase : array[TNumMode] of word = (10,16,8,2);
  ByteDigits : array[TNumMode] of word = (3,2,3,8);    // Stellen pro Byte

type
  TNumUpDown = class(TWinControl)
  private
    FArrowKeys: Boolean;
    FAssociate: TWinControl;
    FDigits      : word;
    FBase : integer;
    FMin: integer;
    FMax: integer;
    FIncrement: Integer;
    FAccelArray : TAccelArray;
    FNewValue: integer;
    FNewValueDelta: integer;
    FPosition: integer;
    FThousands: Boolean;
    FWrap: Boolean;
    FOnClick: TUDClickEvent;
    FAlignButton: TUDAlignButton;
    FOrientation: TUDOrientation;
    FOnChanging: TUDChangingEvent;
    FOnChangingEx: TUDChangingEventEx;
    procedure UndoAutoResizing(Value: TWinControl);
    procedure SetAssociate(Value: TWinControl);
    function GetPosition: integer;
    procedure SetDigits(Value: word);
    function GetHex : boolean;
    procedure SetHex(Value: boolean);
    procedure SetMin(Value: integer);
    procedure SetMax(Value: integer);
    function SetAccelArray : TAccelArray;
    procedure SetIncrement(Value: Integer);
    procedure SetPosition(Value: integer);
    procedure SetAlignButton(Value: TUDAlignButton);
    procedure SetOrientation(Value: TUDOrientation);
    procedure SetArrowKeys(Value: Boolean);
    procedure SetThousands(Value: Boolean);
    procedure SetWrap(Value: Boolean);
    function CheckValue (NewValue: integer): integer;
    procedure CMAllChildrenFlipped(var Message: TMessage); message CM_ALLCHILDRENFLIPPED;
    procedure CNNotify(var Message: TWMNotify); message CN_NOTIFY;
    procedure WMHScroll(var Message: TWMHScroll); message CN_HSCROLL;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure WMVScroll(var Message: TWMVScroll); message CN_VSCROLL;
  protected
    function DoCanChange(NewVal: integer; Delta: integer): Boolean;
    function CanChange: Boolean; dynamic;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Click(Button: TUDBtnType); reintroduce; dynamic;
  public
    constructor Create(AOwner: TComponent); override;
    function NumToStr(Value : integer) : string;
  published
    property AlignButton: TUDAlignButton read FAlignButton write SetAlignButton default udRight;
    property ArrowKeys: Boolean read FArrowKeys write SetArrowKeys default True;
    property Associate: TWinControl read FAssociate write SetAssociate;
    property Digits : word read FDigits write SetDigits;
    property Hex : boolean read GetHex write SetHex default false;
    property Min: integer read FMin write SetMin default 0;
    property Max: integer read FMax write SetMax default 100;
    property Increment: Integer read FIncrement write SetIncrement default 1;
    property Orientation: TUDOrientation read FOrientation write SetOrientation default udVertical;
    property Position: integer read GetPosition write SetPosition default 0;
    property Thousands: Boolean read FThousands write SetThousands default True;
    property Wrap: Boolean read FWrap write SetWrap default False;
    property OnChanging: TUDChangingEvent read FOnChanging write FOnChanging;
    property OnChangingEx: TUDChangingEventEx read FOnChangingEx write FOnChangingEx;
    property OnClick: TUDClickEvent read FOnClick write FOnClick;
    property Anchors;
    property Enabled;
    property Hint;
    property Constraints;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnContextPopup;
    property OnEnter;
    property OnExit;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

  TNumberEdit = class(TCustomEdit)
  private
    FBits,                   // Anzahl der Bits
    FDigits,                 // Anzahl der Anzeigestellen
    FGroupDigits : word;     // Gruppieren der Stellen
    FPrefix      : boolean;  // Prefix ! = bin, % = Oktal, $ = hex
    FBytes       : boolean;  // Grupp
    FNumMode     : TNumMode;
    FAlignment   : TAlignment;
    procedure SetAlignment(Value: TAlignment);
    procedure SetBits(Value : word);
    procedure SetDigits(Value: word);
    procedure SetNumMode(Value: TNumMode);
    function GetValue : int64;
    procedure SetValue(Value : int64);
    procedure SetPrefix (Value : boolean);
    procedure SetBytes(Value : boolean);
    procedure SetGroupDigits(Value : word);
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure CMExit(var Message: TCMGotFocus); message CM_EXIT;
    procedure WMPaste(var Message: TWMPaste);   message WM_PASTE;
    procedure WMCut(var Message: TWMCut);   message WM_CUT;
  protected
    FValue       : int64;
    function IsValidChar(Key: Char): Boolean; virtual;
    procedure KeyPress(var Key: Char); override;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    { Standard-Eigenschaften fr TEDIT verffentlichen }
    property AutoSelect;
    property BorderStyle;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    { Eigenschaften von NumberEdit }
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property Bits : word read FBits write SetBits default 32;
    property Digits : word read FDigits write SetDigits default 0;
    property Prefix : boolean read FPrefix write SetPrefix default true;
    property Bytes : boolean read FBytes write SetBytes default false;
    property GroupDigits : word read FGroupDigits write SetGroupDigits default 0;
    property NumMode : TNumMode read FNumMode write SetNumMode default nmDecimal;
    property Value : int64 read GetValue write SetValue;
  end;

  TRangeEdit = class (TNumberEdit)
  private
    FMinValue,
    FMaxValue   : longint;
    FRangeCheck : boolean;
    function GetValue : longint;
    procedure SetValue(i : integer);
    procedure SetMinValue(i : integer);
    procedure SetMaxValue(i : integer);
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
  public
    constructor Create(AOwner: TComponent); override;
    function CheckValue (NewValue: LongInt): LongInt;
  published
    { Published declarations }
    property MinValue : longint read FMinValue write SetMinValue;
    property MaxValue : longint read FMaxValue write SetMaxValue;
    property RangeCheck : boolean read FRangeCheck write FRangeCheck default true;
    property Value : longint read GetValue write SetValue;
  end;

  TFloatEdit = class(TCustomEdit)
  private
    FDigits,FDecimal  : word;
    FFloatFormat      : TNumFloatFormat;
    FAlignment        : TAlignment;
    procedure SetAlignment(Value: TAlignment);
    procedure SetDigits(Value: word);
    procedure SetDecimal(Value: word);
    procedure SetFloatFormat(Value: TNumFloatFormat);
    function GetValue : double;
    procedure SetValue(x : double);
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure CMExit(var Message: TCMGotFocus); message CM_EXIT;
    procedure WMPaste(var Message: TWMPaste);   message WM_PASTE;
    procedure WMCut(var Message: TWMCut);   message WM_CUT;
  protected
    FValue            : double;
    function IsValidChar(Key: Char): Boolean; virtual;
    procedure KeyPress(var Key: Char); override;
    procedure CreateParams(var Params: TCreateParams); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    { Standard-Eigenschaften fr TEDIT verffentlichen }
    property AutoSelect;
    property AutoSize;
    property BorderStyle;
    property CharCase;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property HideSelection;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ReadOnly;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    { neue Eigenschaften von FloatEdit }
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
    property Digits : word read FDigits write SetDigits;
    property Decimal: word read FDecimal write SetDecimal;
    property FloatFormat : TNumFloatFormat read FFloatFormat write SetFloatFormat;
    property Value : double read GetValue write SetValue;
  end;

  TFloatRangeEdit = class (TFloatEdit)
  private
    FMinValue,
    FMaxValue   : double;
    FRangeCheck : boolean;
    function GetValue : double;
    procedure SetValue(x : double);
    procedure SetMinValue(x : double);
    procedure SetMaxValue(x : double);
    procedure CMExit(var Message: TCMExit);   message CM_EXIT;
  public
    constructor Create(AOwner: TComponent); override;
    function CheckValue (NewValue: double): double;
  published
    { Published declarations }
    property MinValue : double read FMinValue write SetMinValue;
    property MaxValue : double read FMaxValue write SetMaxValue;
    property RangeCheck : boolean read FRangeCheck write FRangeCheck default true;
    property Value : double read GetValue write SetValue;
  end;

  TFloatComboBox = class(TCustomComboBox)
  private
    FDigits,FDecimal  : word;
    FFloatFormat      : TNumFloatFormat;
    procedure SetDigits(Value: word);
    procedure SetDecimal(Value: word);
    procedure SetFloatFormat(Value: TNumFloatFormat);
    function GetValue : double;
    procedure SetValue(x : double);
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure CMExit(var Message: TCMGotFocus); message CM_EXIT;
    procedure WMPaste(var Message: TWMPaste);   message WM_PASTE;
    procedure WMCut(var Message: TWMCut);   message WM_CUT;
  protected
    FValue       : double;
    function IsValidChar(Key: Char): Boolean; virtual;
    procedure KeyPress(var Key: Char); override;
  public
    constructor Create(AOwner: TComponent); override;
    function ValToStr(Value : double) : string; virtual;
    procedure AddValue (Value : double);
  published
    property Style; {Must be published before Items}
    property Color;
    property Ctl3D;
    property DragMode;
    property DragCursor;
    property DropDownCount;
    property Enabled;
    property Font;
    property ItemHeight;
    property Items;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Sorted;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawItem;
    property OnDropDown;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMeasureItem;
    { neue Eigenschaften von FloatComboBox }
    property Digits : word read FDigits write SetDigits;
    property Decimal: word read FDecimal write SetDecimal;
    property FloatFormat : TNumFloatFormat read FFloatFormat write SetFloatFormat;
    property Value : double read GetValue write SetValue;
  end;

  TDegreeComboBox = class(TFloatComboBox)
  private
    FFormat : TAngleFormat;
    function GetValue : double;
    procedure SetValue(x : double);
    procedure CMExit(var Message: TCMExit);   message CM_EXIT;
  protected
    function IsValidChar(Key: Char): Boolean; override;
  public
    function ValToStr(Value : double) : string; override;
  published
    property Value : double read GetValue write SetValue;
    property Format : TAngleFormat read FFormat write FFormat default fmDezMin;
  end;

  EUserRangeError = class (EIntError)
  public
    constructor Create (Min,Max : longint;
                        NumMode : TNumMode);
  end;

  EUserFloatRangeError = class (EIntError)
  public
    constructor Create (Min,Max          : double;
                        FloatFormat      : TNumFloatFormat;
                        Precision,Digits : word);
  end;

function FloatToStrF(Value: Extended; Format: TNumFloatFormat;
  Precision, Digits: Integer): string; overload;

procedure Register;

(* ----------------------------------------------------------------------- *)
implementation

uses ComStrs, JrUtils;

{------------------------------------------------------------------}
function FloatToStrF(Value: Extended; Format: TNumFloatFormat;
  Precision, Digits: Integer): string;
begin
  if Format=ffNormalized then Result:=FloatToStrP(Value,DecimalSeparator,Precision)
  else Result:=FloatToStrF(Value,TFloatFormat(Format),Precision,Digits);
  end;

{------------------------------------------------------------------}
(* Winkel-Wert konvertieren *)
function StrToDeg (s     : string;
                   var v : double) : boolean;
var
 t      : String;
 x,y    : double;
 j,k    : integer;
begin
  j:=0;
  k:=pos(',',s); if k>0 then s[k]:='.';
  if length(s)<>0 then begin
    k:=pos(':',s);
    if k=0 then k:=pos('/',s);
    if k=0 then k:=pos('"',s);
    if k=0 then val(s,v,j)
    else begin (* Grad:Min *)
      val (copy(s,1,pred(k)),x,j); (* Grad *)
      if j=0 then begin
        t:=copy (s,succ(k),255);
        k:=pos(':',t);
        if k=0 then k:=pos('/',t);
        if k=0 then k:=pos(#$27,s);
        if k=0 then begin (* dez.Min *)
          if length(t)>0 then begin
            val (t,v,j);
            if (j=0) and (v<60.0) then begin
              if x>=0.0 then v:=x+v/60.0
              else v:=x-v/60.0;
              end
            else j:=1;
            end
          else v:=x;
          end
        else begin (* Sek. *)
          val (copy (t,1,pred(k)),y,j);  (* Min *)
          if (j=0) and (y<60.0) then begin
            t:=copy (t,succ(k),255);
            if length(t)>0 then begin
              val (t,v,j); (* Sek. *)
              if (j=0) and (v<60.0) then begin
                if x>=0.0 then v:=x+(y+v/60.0)/60.0
                else v:=x-(y-v/60.0)/60.0;
                end
              else j:=1;
              end
            else begin
              if x>=0.0 then v:=x+y/60.0
              else v:=x-y/60.0;
              end;
            end
          else j:=1;
          end;
        end;
      end;
    end
  else v:=0.0;
  StrToDeg:=j=0;
  end;

{------------------------------------------------------------------}
(* Winkel in String mit vorgebenem Format umsetzen
   Stellen: w : gesamt
            d : hinter Komma (einschl. Minuten und Sekunden
   Typ : m = fmDezGrad  : Grad,xxx
             fmDezMin   : Grad:Minuten,xx
             fmSekunden : Grad:Minuten:Sekunden,xxx *)
function DegToStr(v      : double;
                  Format : TAngleFormat;
                  w,d    : integer) : string;
var
  s,t   : String;
  x,y,z : extended;
  i     : integer;
begin
  x:=int(abs(v)); y:=60.0*frac(abs(v));
  case Format of
  fmDezGrad : str(v:w:d,s);
  fmDezMin : begin (* Grad:Min *)
      str (y:(w-6):(d-2),t);
      if copy(t,1,2)='60' then begin
        y:=0.0; x:=x+1.0;
        str (y:(w-6):(d-2),t);
        end;
      if v<0.0 then x:=-x;
      str(x:5:0,s); s:=s+':'+t;
      end;
  fmSekunden : begin (* Grad:Min:Sek *)
      z:=int(y); y:=60.0*frac(y);
      str(y:(w-8):(d-4),t);
      if copy(t,1,2)='60' then begin
        y:=0.0; z:=z+1.0;
        if z>=60.0 then begin
          z:=0.0; x:=x+1.0;
          end;
        str (y:(w-8):(d-4),t);
        end;
      str (z:2:0,s); t:=s+':'+t;
      if v<0.0 then x:=-x;
      str(x:4:0,s); s:=s+':'+t;
      end;
    end;
  i:=pos('.',s); if i>0 then s[i]:=',';
  DegToStr:=s;
  end;

(* ----------------------------------------------------------------------- *)
(* Fehlermeldungen generieren *)
constructor EUserRangeError.Create (Min,Max : longint;
                                    NumMode : TNumMode);
var
  s : string;
begin
  case NumMode of
  nmHex   : s:='$'+IntToHex(Min,1)+' , $'+IntToHex(Max,1)+')';
  nmOctal : s:='%'+IntToOctal(Min,1,0)+' , %'+IntToOctal(Max,1,0)+')';
  nmBin   : s:='!'+IntToBin(Min,1,0)+' , !'+IntToBin(Max,1,0)+')';
    else s:=IntToStr(Min)+' .. '+IntToStr(Max)+')';
    end;
  inherited Create ('Fehler:'#13'zulssiger Bereich: ('+s);
  end;

constructor EUserFloatRangeError.Create (Min,Max          : double;
                                         FloatFormat      : TNumFloatFormat;
                                         Precision,Digits : word);
var
  s : string;
begin
  if FloatFormat=ffNormalized then
    s:=FloatToStrP(Min,DecimalSeparator,Precision)+
                       ' .. '+FloatToStrP(Max,DecimalSeparator,Precision)+')'
  else s:=FloatToStrF(Min,TFloatFormat(FloatFormat),Precision,Digits)+' .. '+
     FloatToStrF(Max,TFloatFormat(FloatFormat),Precision,Digits)+')';
  inherited Create ('Fehler:'#13'zulssiger Bereich: ('+s);
  end;

(* ----------------------------------------------------------------------- *)
{ TUpDown }

constructor TNumUpDown.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Width := GetSystemMetrics(SM_CXVSCROLL);
  Height := GetSystemMetrics(SM_CYVSCROLL);
  Height := Height + (Height div 2);
  FArrowKeys := True;
  FWrap := false;
  FPosition := 0;
  FMin := 0;
  FMax := 100;
  FIncrement := 1;
  FAccelArray:=SetAccelArray;
  FDigits:=4;
  FBase:=10;
  FAlignButton := udRight;
  FOrientation := udVertical;
  FThousands := True;
  ControlStyle := ControlStyle - [csDoubleClicks];
end;

procedure TNumUpDown.CreateParams(var Params: TCreateParams);
begin
  InitCommonControl(ICC_UPDOWN_CLASS);
  inherited CreateParams(Params);
  with Params do
  begin
//    Style := Style or UDS_SETBUDDYINT;
    if FAlignButton = udRight then Style := Style or UDS_ALIGNRIGHT
    else Style := Style or UDS_ALIGNLEFT;
    if FOrientation = udHorizontal then Style := Style or UDS_HORZ;
    if FArrowKeys then Style := Style or UDS_ARROWKEYS;
    if not FThousands then Style := Style or UDS_NOTHOUSANDS;
    if FWrap then Style := Style or UDS_WRAP;
  end;
  CreateSubClass(Params, UPDOWN_CLASS);
  with Params.WindowClass do
    style := style and not (CS_HREDRAW or CS_VREDRAW) or CS_DBLCLKS;
end;

procedure TNumUpDown.CreateWnd;
var
  OrigWidth: Integer;
begin
  OrigWidth := Width;  { control resizes width - disallowing user to set width }
  inherited CreateWnd;
  if FAssociate <> nil then
  begin
    UndoAutoResizing(FAssociate);
    SendMessage(Handle, UDM_SETBUDDY, FAssociate.Handle, 0);
  end;
  Width := OrigWidth;
  SendMessage(Handle, UDM_SETBASE, FBase, 0);
  SendMessage(Handle, UDM_SETRANGE32, FMin, FMax);
  SendMessage(Handle, UDM_SETPOS32, 0, FPosition);
//  SendMessage(Handle, UDM_GETACCEL, NumAccel, Longint(@AccelArray));
//  AccelArray[0].nInc := FIncrement;
  SendMessage(Handle, UDM_SETACCEL, NumAccel, Longint(@FAccelArray));
  if FAssociate is TCustomEdit then
    TCustomEdit(FAssociate).Text := NumToStr(FPosition);
end;

function TNumUpDown.CheckValue (NewValue: integer): integer;
begin
  Result:=NewValue;
  if (FMax<>FMin) then begin
    if NewValue<FMin then Result:=FMin
    else if NewValue>FMax then Result:=FMax;
    end;
  end;

function TNumUpDown.SetAccelArray : TAccelArray;
var
  i,n : integer;
begin
  n:=FIncrement;
  for i:=0 to NumAccel-1 do with Result[i] do begin
    nSec:=3*i;
    nInc:=n;
    n:=10*n;
    end;
  end;

procedure TNumUpDown.WMVScroll(var Message: TWMVScroll);
begin
  inherited;
  if Message.ScrollCode = SB_THUMBPOSITION then
  begin
    if Message.Pos > FPosition then
      Click(btNext)
    else if Message.Pos < FPosition then Click(btPrev);

    FPosition := Message.Pos;
  end;
end;

procedure TNumUpDown.WMSize(var Message: TWMSize);
var
  R: TRect;
begin
  inherited;
  R := ClientRect;
  InvalidateRect(Handle, @R, False);
end;

procedure TNumUpDown.WMHScroll(var Message: TWMHScroll);
begin
  inherited;
  if Message.ScrollCode = SB_THUMBPOSITION then
  begin
    if Message.Pos > FPosition then Click(btNext)
    else if Message.Pos < FPosition then Click(btPrev);
    FPosition := Message.Pos;
  end;
end;

function TNumUpDown.DoCanChange(NewVal: integer; Delta: integer): Boolean;
begin
  FNewValue := CheckValue(NewVal);
  FNewValueDelta := Delta;

  Result := CanChange;
end;

function TNumUpDown.CanChange: Boolean;
var
  Direction: TUpDownDirection;
begin
  Result := True;
  Direction := updNone;

  if (FNewValue < Min) and (FNewValueDelta < 0) or
  (FNewValue > Max) and (FNewValueDelta > 0) then
    Direction := updNone
  else if FNewValueDelta < 0 then
    Direction := updDown
  else if FNewValueDelta > 0 then
    Direction := updUp;
  if FAssociate is TCustomEdit then
    TCustomEdit(FAssociate).Text := NumToStr(FNewValue);

  if Assigned(FOnChanging) then
    FOnChanging(Self, Result);
  if Assigned(FOnChangingEx) then
    FOnChangingEx(Self, Result, FNewValue, Direction);
end;

procedure TNumUpDown.CMAllChildrenFlipped(var Message: TMessage);
begin
  if FAlignButton = udRight then
    SetAlignButton(udLeft)
  else
    SetAlignButton(udRight);
end;

procedure TNumUpDown.CNNotify(var Message: TWMNotify);
begin
  with Message do
    if NMHdr^.code = UDN_DELTAPOS then
    begin
      LongBool(Result) := not DoCanChange(PNMUpDown(NMHdr).iPos + PNMUpDown(NMHdr).iDelta,
                                          PNMUpDown(NMHdr).iDelta);
    end;
  end;

procedure TNumUpDown.Click(Button: TUDBtnType);
begin
  if Assigned(FOnClick) then FOnClick(Self, Button);
  end;

function TNumUpDown.NumToStr(Value : integer) : string;
begin
  if FBase<>10 then Result:='$'+IntToHex(Value,Digits)
  else Result:=IntToStr(Value);
  end;

procedure TNumUpDown.SetAssociate(Value: TWinControl);
var
  I: Integer;

  function IsClass(ClassType: TClass; const Name: string): Boolean;
  begin
    Result := True;
    while ClassType <> nil do
    begin
      if ClassType.ClassNameIs(Name) then Exit;
      ClassType := ClassType.ClassParent;
    end;
    Result := False;
  end;

begin
  if Value <> nil then
    for I := 0 to Parent.ControlCount - 1 do // is control already associated
      if (Parent.Controls[I] is TNumUpDown) and (Parent.Controls[I] <> Self) then
        if TNumUpDown(Parent.Controls[I]).Associate = Value then
          raise Exception.CreateResFmt(@sUDAssociated,
            [Value.Name, Parent.Controls[I].Name]);

  if FAssociate <> nil then { undo the current associate control }
  begin
    if HandleAllocated then
      SendMessage(Handle, UDM_SETBUDDY, 0, 0);
    FAssociate := nil;
  end;

  if (Value <> nil) and (Value.Parent = Self.Parent) and
    not (Value is TNumUpDown) and
    not (Value is TCustomTreeView) and not (Value is TCustomListView) and
    not IsClass(Value.ClassType, 'TDBEdit') and
    not IsClass(Value.ClassType, 'TDBMemo') then
  begin
    if HandleAllocated then
    begin
      UndoAutoResizing(Value);
      SendMessage(Handle, UDM_SETBUDDY, Value.Handle, 0);
    end;
    FAssociate := Value;
    if Value is TCustomEdit then
      TCustomEdit(Value).Text := NumToStr(FPosition);
  end;
end;

procedure TNumUpDown.UndoAutoResizing(Value: TWinControl);
var
  OrigWidth, NewWidth, DeltaWidth: Integer;
  OrigLeft, NewLeft, DeltaLeft: Integer;
begin
  { undo Window's auto-resizing }
  OrigWidth := Value.Width;
  OrigLeft := Value.Left;
  SendMessage(Handle, UDM_SETBUDDY, Value.Handle, 0);
  NewWidth := Value.Width;
  NewLeft := Value.Left;
  DeltaWidth := OrigWidth - NewWidth;
  DeltaLeft := NewLeft - OrigLeft;
  Value.Width := OrigWidth + DeltaWidth;
  Value.Left := OrigLeft - DeltaLeft;
end;

procedure TNumUpDown.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FAssociate) then
    if HandleAllocated then
    begin
      SendMessage(Handle, UDM_SETBUDDY, 0, 0);
      FAssociate := nil;
    end;
end;

function TNumUpDown.GetPosition: integer;
begin
  if HandleAllocated then
  begin
    Result := SendMessage(Handle, UDM_GETPOS32, 0, 0);
    FPosition := Result;
  end
  else Result := FPosition;
end;

procedure TNumUpDown.SetMin(Value: integer);
begin
  if Value <> FMin then
  begin
    FMin := Value;
    if HandleAllocated then
      SendMessage(Handle, UDM_SETRANGE32, FMin, FMax);
  end;
end;

procedure TNumUpDown.SetDigits(Value: word);
begin
  if FDigits <> Value then begin
    FDigits := Value;
    SetPosition(FPosition);
    RecreateWnd;
    end;
  end;

function TNumUpDown.GetHex : boolean;
begin
  Result:=FBase<>10;
  end;

procedure TNumUpDown.SetHex(Value: boolean);
begin
  if Value then FBase:=16 else FBase:=10;
  SendMessage(Handle, UDM_SETBASE, FBase, 0);
  RecreateWnd;
  end;

procedure TNumUpDown.SetMax(Value: integer);
begin
  if Value <> FMax then
  begin
    FMax := Value;
    if HandleAllocated then
      SendMessage(Handle, UDM_SETRANGE32, FMin, FMax);
    end;
  end;

procedure TNumUpDown.SetIncrement(Value: Integer);
begin
  if Value <> FIncrement then
  begin
    FIncrement := Value;
    FAccelArray:=SetAccelArray;
    if HandleAllocated then begin
//      SendMessage(Handle, UDM_GETACCEL, 1, Longint(@AccelArray));
//      AccelArray[0].nInc := Value;
      SendMessage(Handle, UDM_SETACCEL, 1, Longint(@FAccelArray));
      end;
    end;
  end;

procedure TNumUpDown.SetPosition(Value: integer);
begin
  if Value <> FPosition then
  begin
    if not (csDesigning in ComponentState) then
      if not DoCanChange(Value, Value-FPosition) then Exit;
    FPosition := Value;
//    if (csDesigning in ComponentState) and (FAssociate <> nil) then
    if FAssociate is TCustomEdit then
        TCustomEdit(FAssociate).Text := NumToStr(FPosition);
    if HandleAllocated then
      SendMessage(Handle, UDM_SETPOS32, 0, FPosition);
    end;
  end;

procedure TNumUpDown.SetOrientation(Value: TUDOrientation);
begin
  if Value <> FOrientation then
  begin
    FOrientation := Value;
    if ComponentState * [csLoading, csUpdating] = [] then
      SetBounds(Left, Top, Height, Width);
    if HandleAllocated then
      SendMessage(Handle, UDM_SETBUDDY, 0, 0);
    RecreateWnd;
  end;
end;

procedure TNumUpDown.SetAlignButton(Value: TUDAlignButton);
begin
  if Value <> FAlignButton then
  begin
    FAlignButton := Value;
    if HandleAllocated then
      SendMessage(Handle, UDM_SETBUDDY, 0, 0);
    RecreateWnd;
  end;
end;

procedure TNumUpDown.SetArrowKeys(Value: Boolean);
begin
  if Value <> FArrowKeys then
  begin
    FArrowKeys := Value;
    if HandleAllocated then
      SendMessage(Handle, UDM_SETBUDDY, 0, 0);
    RecreateWnd;
  end;
end;

procedure TNumUpDown.SetThousands(Value: Boolean);
begin
  if Value <> FThousands then
  begin
    FThousands := Value;
    if HandleAllocated then
      SendMessage(Handle, UDM_SETBUDDY, 0, 0);
    RecreateWnd;
  end;
end;

procedure TNumUpDown.SetWrap(Value: Boolean);
begin
  if Value <> FWrap then
  begin
    FWrap := Value;
    if HandleAllocated then
      SendMessage(Handle, UDM_SETBUDDY, 0, 0);
    RecreateWnd;
  end;
end;

(* ----------------------------------------------------------------------- *)
(* Integerzahl eingeben und anzeigen (dezimal oder hex) *)
constructor TNumberEdit.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FNumMode:=nmDecimal; FDigits:=0;
  FBits:=32; FPrefix:=true; FGroupDigits:=0;
  FBytes:=false; FValue:=0;
  Alignment:=taLeftJustify;
  Text:='0';
  end;

procedure TNumberEdit.SetBits(Value : word);
begin
  if Value=0 then Value:=32;
  if FBits<>Value then begin
    FBits:= Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TNumberEdit.SetDigits(Value: word);
begin
  if FDigits<>Value then begin
    FDigits := Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TNumberEdit.SetPrefix (Value : boolean);
begin
  if FPrefix<>Value then begin
    FPrefix:=Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TNumberEdit.SetBytes (Value : boolean);
begin
  if FBytes<>Value then begin
    FBytes:=Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TNumberEdit.SetGroupDigits(Value : word);
begin
  if FGroupDigits<>Value then begin
    FGroupDigits:=Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TNumberEdit.SetNumMode(Value: TNumMode);
begin
  if FNumMode<>Value then begin
    FNumMode:=Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TNumberEdit.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then begin
    FAlignment := Value;
    RecreateWnd;
//    Invalidate;
    end;
  end;

procedure TNumberEdit.CreateParams(var Params: TCreateParams);
const
  Alignments: array[Boolean, TAlignment] of DWORD =
    ((ES_LEFT, ES_RIGHT, ES_CENTER),(ES_RIGHT, ES_LEFT, ES_CENTER));
begin
  inherited CreateParams(Params);
  CreateSubClass(Params, 'EDIT');
  with Params do begin
    Style := Style or Alignments[UseRightToLeftAlignment, FAlignment];
    end;
  end;

function TNumberEdit.GetValue : int64;
var
  s,t : string;
  v   : int64;
  nm : TNumMode;
begin
  t:=Text;
  if length(t)>0 then begin
    if t[1]='$' then nm:=nmHex
    else if t[1]='%' then nm:=nmOctal
    else if t[1]='!' then nm:=nmBin
    else nm:=FNumMode;
    FValue:=0; v:=0;
    repeat
      s:=ReadNxtStr(t,'_');
      try
        case nm of
        nmHex   : v:=HexStrToInt(s);
        nmOctal : v:=OctalStrToInt(s);
        nmBin   : v:=BinStrToInt(s);
          else v:=StrToInt(DelSp(s));
          end;
      except
        on EConvertError do begin
         MessageBeep(0);
         end;
        end;
      FValue:=256*FValue+v;
      until length(t)=0;
    end
  else FValue:=0;
  Result:=FValue;
  end;

procedure TNumberEdit.SetValue(Value : int64);
var
  i,n,m : integer;
  x   : double;
  s   : string;

  function IntToVal (v : int64; n : word) : string;
  begin
    case FNumMode of
    nmHex   : begin
              Result:=IntToHexStr(v,n,FGroupDigits);
              if FPrefix then Result:='$'+Result;
              end;
    nmOctaL : begin
              Result:=IntToOctal(v,n,FGroupDigits);
              if FPrefix then Result:='%'+Result;
              end;
    nmBin   : begin
              Result:=IntToBin(v,n,FGroupDigits);
              if FPrefix then Result:='!'+Result;
              end;
      else if FDigits=0 then Result:=IntToStr(v)
           else Result:=IntToDecimal(v,n,FGroupDigits);
      end;
    end;

begin
  if FValue<>Value then FValue:=Value;
  if FDigits=0 then begin
    x:=pwr2(FBits);
    n:=round(ln(x)/ln(NumBase[FNumMode])+0.5);
    end
  else n:=FDigits;
  if FBytes then begin
    m:=(FBits-1) div 8+1; // Anzahl Bytes
    s:='';
    for i:=1 to m do begin
      if (i>1) then s:='_'+s;
      s:=IntToVal(Value and $FF,ByteDigits[FNumMode])+s;
      Value:=Value div 256;
      end;
    end
  else s:=IntToVal(Value,n);
  Text:=s;
//  FValue:=GetValue;
  end;

procedure TNumberEdit.KeyPress(var Key: Char);
begin
  inherited KeyPress(Key);
  if not IsValidChar(Key) then begin
    Key := #0;
    MessageBeep(0)
    end;
  end;

function TNumberEdit.IsValidChar(Key: Char): Boolean;
begin
{  if length(Text)=0 then begin  // prefix
     if Key='$' then NumMode:=nmHex
     else if Key='%' then NumMode:=nmOctal
     else if Key='!' then NumMode:=nmBin;
    end;
  case NumMode of
  nmHex   : Result := (Key in ['$','0'..'9','A'..'F','a'..'f']);
  nmOctal : Result := (Key in ['%','0'..'7']);
  nmBin   : Result := (Key in ['!','0','1']);
    else Result := (Key in ['+','-','0'..'9']);
    end; }
  Result := (Key in ['$','%','!','+','-','0'..'9','A'..'F','a'..'f']);
  Result:= Result or (Key < #32); // and (Key <> Chr(VK_RETURN)));
  end;

procedure TNumberEdit.CMExit(var Message: TCMGotFocus);
begin
  GetValue;
  end;

procedure TNumberEdit.CMEnter(var Message: TCMGotFocus);
begin
  if AutoSelect and not (csLButtonDown in ControlState) then SelectAll;
  inherited;
  end;

procedure TNumberEdit.WMPaste(var Message: TWMPaste);
begin
  if ReadOnly then Exit;
  inherited;
  end;

procedure TNumberEdit.WMCut(var Message: TWMPaste);
begin
  if ReadOnly then Exit;
  inherited;
  end;

(* ----------------------------------------------------------------------- *)
(* Integerzahl in einem vorgegebenen Bereich (MinValue, MaxValue) eingeben
   und anzeigen (dezimal oder hex) *)
constructor TRangeEdit.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  MinValue:=0; MaxValue:=100;
  RangeCheck:=true; 
  end;

procedure TRangeEdit.SetMinValue(i : integer);
begin
  FMinValue:=i;
  FValue:=CheckValue(FValue);
  end;

procedure TRangeEdit.SetMaxValue(i : integer);
begin
  FMaxValue:=i;
  FValue:=CheckValue(FValue);
  end;

procedure TRangeEdit.SetValue(i : longint);
begin
  if FRangeCheck then i:=CheckValue(i);
  inherited SetValue(i);
  end;

function TRangeEdit.GetValue : longint;
begin
  try
    FValue:=StrToInt(Text);
  except
    on EConvertError do begin
      MessageBeep(0);
      FValue:=FMinValue;
      end;
    end;
  SetValue(FValue);
  Result:=FValue;
  end;

function TRangeEdit.CheckValue (NewValue: LongInt): LongInt;
begin
  Result := NewValue;
  if (FMaxValue <> FMinValue) then begin
    if NewValue < FMinValue then Result := FMinValue
    else if NewValue > FMaxValue then Result := FMaxValue;
    end;
  end;

procedure TRangeEdit.CMExit(var Message: TCMExit);
begin
  inherited;
  if FRangeCheck and (CheckValue(Value)<>Value) then begin
    SetValue (Value);
    MessageBeep(0);
    SetFocus;
    raise EUserRangeError.Create (FMinValue,FMaxValue,FNumMode);
    end
  end;

(* ----------------------------------------------------------------------- *)
(* Fliekommazahl eingeben und anzeigen *)
constructor TFloatEdit.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FloatFormat:=ffFixed;
  Digits:=10; Decimal:=2;
  Alignment:=taLeftJustify;
  Value:=0;
  end;

procedure TFloatEdit.SetAlignment(Value: TAlignment);
begin
  if FAlignment <> Value then begin
    FAlignment := Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TFloatEdit.SetDigits(Value: word);
begin
  if FDigits <> Value then begin
    FDigits := Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TFloatEdit.SetDecimal(Value: word);
begin
  if FDecimal <> Value then begin
    FDecimal := Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TFloatEdit.SetFloatFormat(Value: TNumFloatFormat);
begin
  if FFloatFormat<> Value then begin
    FFloatFormat:=Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TFloatEdit.CreateParams(var Params: TCreateParams);
const
  Alignments: array[Boolean, TAlignment] of DWORD =
    ((ES_LEFT, ES_RIGHT, ES_CENTER),(ES_RIGHT, ES_LEFT, ES_CENTER));
begin
  inherited CreateParams(Params);
  CreateSubClass(Params, 'EDIT');
  with Params do begin
    Style := Style or Alignments[UseRightToLeftAlignment, FAlignment];
    end;
  end;

function TFloatEdit.GetValue : double;
begin
  try
    FValue:=StrToFloat(Text);
  except
    on EConvertError do FValue:=0;
    end;
  Result:=FValue;
  end;

procedure TFloatEdit.SetValue(x : double);
begin
  FValue:=x;
  if FloatFormat=ffNormalized then Text:=FloatToStrP(x,DecimalSeparator,FDigits)
  else Text:=FloatToStrF(x,TFloatFormat(FFloatFormat),FDigits,FDecimal);
  end;

procedure TFloatEdit.KeyPress(var Key: Char);
begin
  if not IsValidChar(Key) then begin
    Key := #0;
    MessageBeep(0)
    end;
  if Key <> #0 then inherited KeyPress(Key);
end;

function TFloatEdit.IsValidChar(Key: Char): Boolean;
begin
  Result := (Key in [DecimalSeparator,'+','-','0'..'9','E','e']);
  Result:= Result or (Key < #32); // and (Key <> Chr(VK_RETURN)));
end;

procedure TFloatEdit.CMExit(var Message: TCMGotFocus);
begin
  GetValue;
  end;

procedure TFloatEdit.CMEnter(var Message: TCMGotFocus);
begin
  if AutoSelect and not (csLButtonDown in ControlState) then SelectAll;
  inherited;
end;

procedure TFloatEdit.WMPaste(var Message: TWMPaste);
begin
  if ReadOnly then Exit;
  inherited;
end;

procedure TFloatEdit.WMCut(var Message: TWMPaste);
begin
  if ReadOnly then Exit;
  inherited;
end;

(* ----------------------------------------------------------------------- *)
(* Fliekommazahl in einem vorgegebenen Bereich (MinValue, MaxValue) eingeben
   und anzeigen *)
constructor TFloatRangeEdit.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  MinValue:=0; MaxValue:=100;
  RangeCheck:=true;
  end;

procedure TFloatRangeEdit.SetMinValue(x : double);
begin
  FMinValue:=x;
  FValue:=CheckValue(FValue);
  end;

procedure TFloatRangeEdit.SetMaxValue(x : double);
begin
  FMaxValue:=x;
  FValue:=CheckValue(FValue);
  end;

procedure TFloatRangeEdit.SetValue(x : double);
begin
  if FRangeCheck then FValue:=CheckValue(x);
  inherited SetValue(x);
  end;

function TFloatRangeEdit.GetValue : double;
begin
  try
    FValue:=StrToFloat(Text);
  except
    on EConvertError do begin
      MessageBeep(0);
      FValue:=FMinValue;
      end;
    end;
  SetValue(FValue);
  Result:=FValue;
  end;

function TFloatRangeEdit.CheckValue (NewValue: double): double;
begin
  Result := NewValue;
  if (FMaxValue <> FMinValue) then begin
    if NewValue < FMinValue then Result := FMinValue
    else if NewValue > FMaxValue then Result := FMaxValue;
    end;
  end;

procedure TFloatRangeEdit.CMExit(var Message: TCMExit);
var
  x : double;
begin
  x:=CheckValue (Value);
  if FRangeCheck and (x<>Value) then begin
    SetValue (x);
    MessageBeep(0);
    SetFocus;
    raise EUserFloatRangeError.Create (FMinValue,FMaxValue,FFloatFormat,FDigits,FDecimal);
    end
  else inherited;
  end;

(* ----------------------------------------------------------------------- *)
(* Fliekommazahl in ComboBox eingeben und anzeigen *)
constructor TFloatComboBox.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FloatFormat:=ffFixed;
  Digits:=10; Decimal:=2;
  Value:=0;
  end;

procedure TFloatComboBox.SetDigits(Value: word);
begin
  if FDigits <> Value then begin
    FDigits := Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TFloatComboBox.SetDecimal(Value: word);
begin
  if FDecimal <> Value then begin
    FDecimal := Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

procedure TFloatComboBox.SetFloatFormat(Value: TNumFloatFormat);
begin
  if FFloatFormat<> Value then begin
    FFloatFormat:=Value;
    SetValue(FValue);
    RecreateWnd;
    end;
  end;

function TFloatComboBox.GetValue : double;
begin
  try
    FValue:=StrToFloat(Text);
  except
    on EConvertError do FValue:=0;
    end;
  Result:=FValue;
  end;

function TFloatComboBox.ValToStr(Value : double) : string;
begin
  if FloatFormat=ffNormalized then Result:=FloatToStrP(Value,DecimalSeparator,FDigits)
  else Result:=FloatToStrF(Value,FFloatFormat,FDigits,FDecimal);
  end;

procedure TFloatComboBox.SetValue(x : double);
begin
  FValue:=x; Text:=ValToStr(x);
  end;

procedure TFloatComboBox.AddValue (Value : double);
begin
  Items.Add(ValToStr(Value));
  end;

procedure TFloatComboBox.KeyPress(var Key: Char);
begin
  if not IsValidChar(Key) then begin
    Key := #0;
    MessageBeep(0)
  end;
  if Key <> #0 then inherited KeyPress(Key);
end;

function TFloatComboBox.IsValidChar(Key: Char): Boolean;
begin
  Result := (Key in [DecimalSeparator,'+','-','0'..'9','E','e']);
  Result:= Result or (Key < #32); // and (Key <> Chr(VK_RETURN)));
end;

procedure TFloatComboBox.CMExit(var Message: TCMGotFocus);
begin
  GetValue;
  end;

procedure TFloatComboBox.CMEnter(var Message: TCMGotFocus);
begin
  if (Style < csDropDownList) and not (csLButtonDown in ControlState) then
    SelectAll;
  inherited;
end;

procedure TFloatComboBox.WMPaste(var Message: TWMPaste);
begin
  if Style >= csDropDownList then Exit;
  inherited;
end;

procedure TFloatComboBox.WMCut(var Message: TWMPaste);
begin
  if Style >= csDropDownList then Exit;
  inherited;
end;

(* ----------------------------------------------------------------------- *)
(* Gradzahl in ComboBox eingeben und anzeigen *)
function TDegreeComboBox.GetValue : double;
var
  val : double;
begin
  if StrToDeg (Text,val) then begin
    Result:=Val; Text:=DegToStr(val,FFormat,FDigits,FDecimal);
    end
  else begin
    MessageDlg ('Fehler in Winkelangabe',mtError,[mbOK],0);
    Result:=0;
    end;
  end;

function TDegreeComboBox.ValToStr(Value : double) : string;
begin
  Result:=DegToStr(Value,FFormat,FDigits,FDecimal);
  end;

procedure TDegreeComboBox.SetValue(x : double);
begin
  FValue:=x;
  Text:=ValToStr(x);
  end;

function TDegreeComboBox.IsValidChar(Key: Char): Boolean;
begin
  Result := (Key in [DecimalSeparator,'+','-','0'..'9',':','/']);
  Result:= Result or (Key < #32); // and (Key <> Chr(VK_RETURN)));
end;

procedure TDegreeComboBox.CMExit(var Message: TCMExit);
begin
  end;

{ ---------------------------------------------------------------- }
procedure Register;
begin
  RegisterComponents(CompPalPage, [TNumUpDown]);
  RegisterComponents(CompPalPage, [TNumberEdit]);
  RegisterComponents(CompPalPage, [TRangeEdit]);
  RegisterComponents(CompPalPage, [TFloatEdit]);
  RegisterComponents(CompPalPage, [TFloatRangeEdit]);
  RegisterComponents(CompPalPage, [TFloatComboBox]);
  RegisterComponents(CompPalPage, [TDegreeComboBox]);
  end;

end.
