(* Kippschalter-Komponenten
   - einfacher Kippschalter (an - aus):  TSwitch
   - Momentanschalter                 : TMomSwitch
   - Zweiseitiger Momentanschalter    : TBiMomSwitch

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Mrz. 2007
   *)

unit Switches;

interface

uses
  SysUtils, Classes, Controls, Windows, Messages, Graphics;

// Definiere CompPalPage (siehe Register)
{$Include UserComps.pas }

type
  TSwitchState = (ssOn,ssOff);
  TSwitchShape = (ssRect,ssRound);
  TSwitchPosition = (spDown,spCenter,spUp);

  TCustomSwitch = class(TCustomControl)
  private
    procedure DoEnter; override;
    procedure DoExit; override;
  protected
    FOuterColor,
    FInnerColor,
    FCenterColor,
    FDownColor,
    FUpColor  : TColor;
    FShape : TSwitchShape;
    FPosition : TSwitchPosition;
    procedure SetOuterColor (Value : TColor);
    procedure SetInnerColor (Value : TColor);
    procedure SetCenterColor (Value : TColor);
    procedure SetDownColor (Value : TColor);
    procedure SetUpColor (Value : TColor);
    procedure SetShape (Value : TSwitchShape);
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property OuterColor : TColor read FOuterColor write SetOuterColor default clSilver;
    property InnerColor : TColor read FInnerColor write SetInnerColor default clGray;
    property CenterColor : TColor read FCenterColor write SetCenterColor default clWhite;
    property DownColor : TColor read FDownColor write SetDownColor default clWhite;
    property UpColor : TColor read FUpColor write SetUpColor default clWhite;
    property Shape : TSwitchShape read FShape write SetShape default ssRound;
  end;

  TSwitch = class(TCustomSwitch)
  private
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
    FState : TSwitchState;
    procedure SetState (Value : TSwitchState);
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property State : TSwitchState read FState write SetState default ssOff;
    property TabOrder;
    property TabStop;
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
  end;

  TMomSwitch = class(TCustomSwitch)
  private
    FOnClick : TNotifyEvent;
    FRestPos : TSwitchPosition;
    procedure SetRestPos (Value : TSwitchPosition);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property RestPosition : TSwitchPosition read FRestPos write SetRestPos default spDown;
    property TabOrder;
    property TabStop;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
  end;

  TBiMomSwitch = class(TCustomSwitch)
  private
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    { Published-Deklarationen }
    property TabOrder;
    property TabStop;
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
  end;

  TGlyphHeight = (gh32,gh48,gh64,gh96);

  TCustomSwitch3D = class(TCustomControl)
  private
    FGlyphHeight : TGlyphHeight;
    FBitMap     : TBitMap;
    FPosition   : TSwitchPosition;
    procedure SetGlyphHeight (Value : TGlyphHeight);
    procedure SetPosition (Value : TSwitchPosition);
    procedure AdjustSize (var W, H: Integer); reintroduce;
    procedure WMSize(var Message: TWMSize);  message WM_SIZE;
  protected
    procedure Paint; override;
    property GlyphHeight : TGlyphHeight read FGlyphHeight write SetGlyphHeight default gh32;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
    end;

  TSwitch3D = class(TCustomSwitch3D)
  private
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
    FState : TSwitchState;
    procedure SetState (Value : TSwitchState);
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property GlyphHeight;
    property State : TSwitchState read FState write SetState default ssOff;
    property TabOrder;
    property TabStop;
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
    end;

  TMomSwitch3D = class(TCustomSwitch3D)
  private
    FOnClick : TNotifyEvent;
    FRestPos : TSwitchPosition;
    procedure SetRestPos (Value : TSwitchPosition);
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    property RestPosition : TSwitchPosition read FRestPos write SetRestPos default spDown;
    property TabOrder;
    property TabStop;
    property OnClick: TNotifyEvent read FOnClick write FOnClick;
  end;

  TBiMomSwitch3D = class(TCustomSwitch3D)
  private
    FOnUpClick: TNotifyEvent;
    FOnDownClick: TNotifyEvent;
  protected
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
  published
    { Published-Deklarationen }
    property TabOrder;
    property TabStop;
    property OnDownClick: TNotifyEvent read FOnDownClick write FOnDownClick;
    property OnUpClick: TNotifyEvent read FOnUpClick write FOnUpClick;
  end;

procedure Register;

implementation

{$R Switch.res}

constructor TCustomSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  { set defaults }
  Width:=16;
  Height:=20;
  FOuterColor:=clSilver;
  FInnerColor:=clGray;
  FCenterColor:=clWhite;
  FDownColor:=clWhite;
  FUpColor:=clWhite;
  FPosition:=spDown;
  end;

procedure TCustomSwitch.SetOuterColor (Value : TColor);
begin
  FOuterColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.SetInnerColor (Value : TColor);
begin
  FInnerColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.SetCenterColor (Value : TColor);
begin
  FCenterColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.SetDownColor (Value : TColor);
begin
  FDownColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.SetUpColor (Value : TColor);
begin
  FUpColor:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.SetShape (Value : TSwitchShape);
begin
  FShape:=Value;
  Invalidate;
  end;

procedure TCustomSwitch.Paint;
var
  n,d,h,m : integer;
begin
  with Canvas do begin
    if Focused then Brush.Color:=FInnerColor
    else Brush.Color:=FOuterColor;
    with Pen do begin
      Color:=clBlack; Width:=0;
      end;
    m:=Height div 10;
    h:=Height-Height div 5;
    Ellipse (0,m,h,m+h);
    if Focused then Brush.Color:=FOuterColor
    else Brush.Color:=FInnerColor;
    n:=h div 8;
    Ellipse (n,m+n,h-n,m+h-n);
    Pen.Width:=n div 2;
    n:=h div 3;
    Brush.Color:=clBlack;
    Ellipse (n,m+n,h-n,m+h-n);
    m:=h div 2;
    n:=Height div 2; d:=Height div 8;
    case FPosition of
    spDown : begin
             Brush.Color:=FDownColor;
             if FShape=ssRect then RoundRect(m-d,n-d,m+d,Height-1,d,d)
             else Ellipse (m-d,n-d,m+d,Height-1);
             end;
    spUp   : begin
             Brush.Color:=FUpColor;
             if FShape=ssRect then RoundRect(m-d,n+d,m+d,1,d,d)
             else Ellipse (m-d,n+d,m+d,0);
             end;
      else begin
         Brush.Color:=clWhite;
         Ellipse (m-d,n-d,m+d,n+d);
        end;
      end;
    end;
  end;

procedure TCustomSwitch.DoEnter;
begin
  inherited;
  Invalidate;
  end;

procedure TCustomSwitch.DoExit; 
begin
  inherited;
  Invalidate;
  end;

//====================================================================
constructor TSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FState:=ssOff;
  FOnUpClick:=nil;
  FOnDownClick:=nil;
  end;

procedure TSwitch.SetState (Value : TSwitchState);
begin
  FState:=Value;
  if FState=ssOn then FPosition:=spUp else FPosition:=spDown;
  Invalidate;
  end;

procedure TSwitch.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    ord('U'),ord('u'),187,VK_ADD: State:=ssOn;
    ord('D'),ord('d'),189,VK_SUBTRACT : State:=ssOff;
    end;
  end;

procedure TSwitch.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if State=ssOff then begin
    State:=ssOn;
    if assigned(FOnUpClick) then FOnUpClick(self);
    end
  else begin
    State:=ssOff;
    if assigned(FOnDownClick) then FOnDownClick(self);
    end;
  end;

//====================================================================
constructor TMomSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPosition:=spDown;
  FOnClick:=nil;
  end;

procedure TMomSwitch.SetRestPos (Value : TSwitchPosition);
begin
  FRestPos:=Value;
  FPosition:=Value;
  Invalidate;
  end;

procedure TMomSwitch.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if FRestPos=spDown then FPosition:=spUp else FPosition:=spDown;
  Invalidate;
  if assigned(FOnClick) then FOnClick(self);
  end;

procedure TMomSwitch.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  FPosition:=FRestPos;
  Invalidate;
  end;

//====================================================================
constructor TBiMomSwitch.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPosition:=spCenter;
  FOnUpClick:=nil;
  FOnDownClick:=nil;
  end;

procedure TBiMomSwitch.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if y>Height div 2 then begin
    FPosition:=spDown;
    if assigned(FOnDownClick) then FOnDownClick(self);
    end
  else begin
    FPosition:=spUp;
    if assigned(FOnUpClick) then FOnUpClick(self);
    end;
  Invalidate;
  end;

procedure TBiMomSwitch.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  FPosition:=spCenter;
  Invalidate;
  end;

//====================================================================
(* 3D Schalter *)
type
  TSwitchSize = record
    GWidth,GHeight : integer;
    end;

const
  GlyphHeights : array [TGlyphHeight] of TSwitchSize =
     ((GWidth:21;GHeight:32),(GWidth:32;GHeight:48),(GWidth:43;GHeight:64),(GWidth:64;GHeight:96));

constructor TCustomSwitch3D.Create (AOwner: TComponent);
begin
  FBitMap:=TBitMap.Create;
  with FBitMap do begin
    Transparent:=True;
    TransparentMode:=tmAuto;
    Height:=32;
    end;
  inherited Create(AOwner);
  GlyphHeight:=gh32;
  FPosition:=spDown;
  end;

destructor TCustomSwitch3D.Destroy;
begin
  FBitMap.Free;
  inherited Destroy;
  end;

procedure TCustomSwitch3D.SetGlyphHeight (Value : TGlyphHeight);
var
  s : string;
begin
  FGlyphHeight:=Value;
  case Value of
  gh48 : if FPosition=spDown then s:='SwitchU48'
         else if FPosition=spUp then s:='SwitchO48'
         else s:='SwitchM48';
  gh64 : if FPosition=spDown then s:='SwitchU64'
         else if FPosition=spUp then s:='SwitchO64'
         else s:='SwitchM64';
  gh96 : if FPosition=spDown then s:='SwitchU96'
         else if FPosition=spUp then s:='SwitchO96'
         else s:='SwitchM96';
    else if FPosition=spDown then s:='SwitchU32'
         else if FPosition=spUp then s:='SwitchO32'
         else s:='SwitchM32';
    end;
  FBitMap.Handle:=LoadBitmap(HInstance,pchar(s));
  with GlyphHeights[Value] do begin
    Width:=GWidth; Height:=GHeight;
    end;
  end;

procedure TCustomSwitch3D.SetPosition (Value : TSwitchPosition);
begin
  FPosition:=Value;
  SetGlyphHeight(FGlyphHeight);
  Invalidate;
  end;

procedure TCustomSwitch3D.AdjustSize(var W,H: Integer);
begin
  if (csLoading in ComponentState) then Exit;
  with FBitMap do begin
    if W<Width then W:=Width;
    if H<Height then H:=Height;
    end;
  end;

procedure TCustomSwitch3D.SetBounds(ALeft,ATop,AWidth,AHeight: Integer);
var
  W, H: Integer;
begin
  W:=AWidth;
  H:=AHeight;
  AdjustSize(W,H);
  inherited SetBounds(ALeft,ATop,W,H);
  end;

procedure TCustomSwitch3D.WMSize(var Message: TWMSize);
var
  W, H: Integer;
begin
  inherited;
  { check for minimum size }
  W:=Width; H:=Height;
  AdjustSize(W,H);
  if (W<>Width) or (H<>Height) then inherited SetBounds(Left,Top,W,H);
  Message.Result:=0;
  end;

procedure TCustomSwitch3D.Paint;
begin
  Canvas.Draw(0,0,FBitMap);
  end;

//====================================================================
constructor TSwitch3D.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  FState:=ssOff;
  FOnUpClick:=nil;
  FOnDownClick:=nil;
  end;

procedure TSwitch3D.SetState (Value : TSwitchState);
begin
  FState:=Value;
  if FState=ssOn then FPosition:=spUp else FPosition:=spDown;
  SetGlyphHeight(FGlyphHeight);
  Invalidate;
  end;

procedure TSwitch3D.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    ord('U'),ord('u'),187,VK_ADD: State:=ssOn;
    ord('D'),ord('d'),189,VK_SUBTRACT : State:=ssOff;
    end;
  end;

procedure TSwitch3D.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if State=ssOff then begin
    State:=ssOn;
    if assigned(FOnUpClick) then FOnUpClick(self);
    end
  else begin
    State:=ssOff;
    if assigned(FOnDownClick) then FOnDownClick(self);
    end;
  end;

//====================================================================
constructor TMomSwitch3D.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetPosition(spDown);
  FOnClick:=nil;
  end;

procedure TMomSwitch3D.SetRestPos (Value : TSwitchPosition);
begin
  FRestPos:=Value;
  FPosition:=Value;
  Invalidate;
  end;

procedure TMomSwitch3D.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if FRestPos=spDown then SetPosition(spUp) else SetPosition(spDown);
  Invalidate;
  if assigned(FOnClick) then FOnClick(self);
  end;

procedure TMomSwitch3D.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  FPosition:=FRestPos;
  Invalidate;
  end;

//====================================================================
constructor TBiMomSwitch3D.Create (AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetPosition(spCenter);
  FOnUpClick:=nil;
  FOnDownClick:=nil;
  end;

procedure TBiMomSwitch3D.MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  if y>Height div 2 then begin
    SetPosition(spDown);
    if assigned(FOnDownClick) then FOnDownClick(self);
    end
  else begin
    SetPosition(spUp);
    if assigned(FOnUpClick) then FOnUpClick(self);
    end;
  Invalidate;
  end;

procedure TBiMomSwitch3D.MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer);
begin
  inherited MouseDown (Button, Shift, X, Y);
  SetPosition(spCenter);
  Invalidate;
  end;

//====================================================================
procedure Register;
begin
  RegisterComponents(CompPalPage, [TSwitch,TMomSwitch,TBiMomSwitch]);
  RegisterComponents(CompPalPage, [TSwitch3D,TMomSwitch3D,TBiMomSwitch3D]);
end;

end.
