{
  Vers. 3.0
  TmMailer with extensions to multipart messaging and attachments with content types
  J. Rathlev, March 2005

  Vers. 3.1
  Added ExtraHeaders
  in multipart messages plaintext is sent first (see RFC 1521 und 2046)

  Vers. 3.2
  works with Indy10    J. Rathlev, 2007-06-07

  Vers. 3.3
  port for SMTP added  J. Rathlev, 2007-11-21

  Vers. 3.4
  TLS support added J. Rathlev, 2008-08-27
    Requires:  Indy10-Versions of libeay32.dll, ssleay32.dll
               SSL certificate (created by OpenSSL)

----------------------------------------------------------
MAS-CompMaker was used to generate this code
MAS-CompMaker, 2000-2002 MASoft, Mats Asplund
----------------------------------------------------------

Component Name: TmMailer
        Author: MASoft, Mats Asplund
      Creation: 2002-08-09
       Version: 2.1
   Description: A mailsender component
        Credit: (place your credits here)
        E-mail: mdp@telia.com
          Site: http://hem.passagen.se/matasp
  Legal issues: All rights reserved 2002 by MASoft, Mats Asplund

Usage:
  This software is provided 'as-is', without any express or
  implied warranty.  In no event will the author be held liable
  for any  damages arising from the use of this software.

  Permission is granted to anyone to use this software for any
  purpose, including commercial applications, and to alter it
  and redistribute it freely, subject to the following
  restrictions:

  1. The origin of this software must not be misrepresented,
     you must not claim that you wrote the original software.
     If you use this software in a product, an acknowledgment
     in the product documentation would be appreciated but is
     not required.

  2. Altered source versions must be plainly marked as such, and
     must not be misrepresented as being the original software.

  3. This notice may not be removed or altered from any source
     distribution.

  4. If you decide to use this software in any of your applications.
     Send me an EMail and tell me about it.

Quick Reference:
  TmMailer inherits from TComponent.

  Key-Properties:
    Body: The Body-part of the EMAil-message.
    Attachments
      AtttachedFiles: Add the files you wan't to attache to this StringList,
      FileName:       or just write filenames here.
    Host: Your Internet provider.
    UserId: Your UserId.
    Password: Your Password.
    Priority: Five levels of priority for the message.
    From: The sender EMail-address.
    Recipients: Add recipients to this list.
    BCCList: BccList is used to store email address of recipients that will
      receive a carbon copy of the message, but do not appear in the
      "To:" headers for the message. One way to prevent addresses to picked up
      by spammer.
    Subject: The messages subject.
    HTML: If true the EMail is marked as HTML else as plain text.
       (Rem. (J. Rathlev): This only works without attachements, see changes below)

  Key-Events:
    OnSendError: Event fired if there is an exception while sending.

  (New in version 2.1)
    OnConnected:    Event fired when client is connected.
    OnDisconnected: -------------"------------ disconnected.

  Key-Methods:
    Send: Sends the message.


--------------------------------------------------------------------------------
Changes by J. Rathlev, Mar. 2005 - Vers. 3
  Added method "TAttachment.Clear";
  Function "Send" returns false on send error
  OnSendError reports e-mail address
  Setting of "ContentType" to attachments ("application/<subtype>" or "<type>/<subtype>")
    (Format of attachment strings: <filename>|<type> e.g. pdf,
     if <type> is omitted content-type has the default value "application/octet-stream")
  Messages with alternative text parts are supported
    (Set HTMLBody to HTML text, Body to plain text)

--------------------------------------------------------------------------------
}
unit mMailer10;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, StdCtrls, ExtCtrls, IdBaseComponent, IdMessage, IdComponent,
  IdTCPConnection, IdTCPClient, IdMessageClient, IdSMTP, IdEMailAddress,
  IdHeaderList, IdAttachmentFile, IdExplicitTLSClientServerBase, IdText;

const
  defHtmlType = 'text/html; charset=iso-8859-1';
  defTextType = 'text/plain; charset=iso-8859-1';
  defEncoding = '8bit';
  defSmtpTimeOut = 15000;

type
  TOnSendError = procedure(Sender: TObject; ErrorMsg: string) of object;
  TGetPassword = procedure(var APassword : string) of object;

  TAttachement = class(TPersistent)
  private
    fAttachedFiles: TStrings;
    fOwner: TComponent;
    procedure SetAttachedFiles(const Value: TStrings);
  public
    constructor Create(Owner: TComponent);
    procedure Clear;
  published
    property AttachedFiles: TStrings read fAttachedFiles write SetAttachedFiles;
  end;

  TmMailer = class(TComponent)
  private
    fAbout: string;
    fHost: string;
    fPort : integer;
    fUserId: string;
    fFrom: TIdEMailAddressItem;
    fRecipients: TIdEMailAddressList;
    fBCC: TIdEMailAddressList;
    fBody,fHTMLBody: TStrings;
    fSubject: string;
    fMsg: TIdMessage;
    fExtraHeaders: TIdHeaderList;
    fSMTP: TIdSMTP;
    fCertPwd,
    fPassword: string;
    fTLSAvail,
    fHTML: Boolean;
    fPriority: TIdMessagePriority;
    fOnSendError: TOnSendError;
    fOnDisconnected: TNotifyEvent;
    fOnConnected: TNotifyEvent;
    fAttachments: TAttachement;
    fGetSSLPassword : TGetPassword;
    fUseTLS : TIdUseTLS;
    procedure SetBody(const Value: TStrings);
    procedure SetHTMLBody(const Value: TStrings);
    procedure IsConnected(Sender: TObject);
    procedure Disconnected(Sender: TObject);
    procedure SetTLS(Value : integer);
    procedure GetCertPassword(var Password : string);
  public
    constructor Create(AOwner: TComponent; const Path,Root,Cert,Key,Pwd : string); overload;
    constructor Create (AOwner : TComponent); overload;
    destructor Destroy; override;
    procedure ClearExtraHeaders;
    procedure AddExtraHeader (fId,fContents : string);
    function Send : boolean;
  published
    property About: string read fAbout;
    property Body: TStrings read fBody write SetBody;
    property HTMLBody: TStrings read fHTMLBody write SetHTMLBody;
    property Attachments: TAttachement read fAttachments write fAttachments;
    property Host: string read fHost write fHost;
    property Port: integer read fPort write fPort;
    property UserId: string read fUserId write fUserId;
    property Password: string read fPassword write fPassword;
    property Priority: TIdMessagePriority read fPriority write fPriority default
      ID_MSG_PRIORITY;
    property From: TIdEMailAddressItem read fFrom write fFrom;
    property Recipients: TIdEMailAddressList read fRecipients write fRecipients;
    property BCCList: TIdEMailAddressList read fBCC write fBCC;
    property Subject: string read fSubject write fSubject;
    property HTML: Boolean read fHTML write fHTML;
    property OnSendError: TOnSendError read fOnSendError write fOnSendError;
    property OnConnected: TNotifyEvent read fOnConnected write fOnConnected;
    property OnDisconnected: TNotifyEvent read fOnDisconnected write fOnDisconnected;
    property OnSSLPassword: TGetPassword read fGetSSLPassword write fGetSSLPassword;
    property SecureTransfer : integer write SetTLS;
  end;

procedure Register;

implementation

uses IdAssignedNumbers, IdSSLOpenSSL, IdIOHandler, IdException, IdResourceStringsProtocols;

{ ------------------------------------------------------------------- }
{ TAttachement }

constructor TAttachement.Create(Owner: TComponent);
begin
  inherited Create;
  fOwner:= Owner;
end;

procedure TAttachement.SetAttachedFiles(const Value: TStrings);
begin
  fAttachedFiles.Assign(Value);
end;

// added Mar. 2004, J. Rathlev
procedure TAttachement.Clear;
begin
  fAttachedFiles.Clear;
  end;

{ ------------------------------------------------------------------- }
{ TmMailer }

constructor TmMailer.Create(AOwner: TComponent; const Path,Root,Cert,Key,Pwd : string);
var
  sr,sc,sk : string;
begin
  inherited Create(AOwner);
  fAttachments:= TAttachement.Create(self);
  fAttachments.fAttachedFiles:= TStringList.Create;
  fPort:=IdPORT_SMTP;
  fBody:= TStringList.Create;
  fHTMLBody:= TStringList.Create;
  fRecipients:= TIdEMailAddressList.Create(Self);
  fBCC:= TIdEMailAddressList.Create(Self);
  fFrom:= TIdEMailAddressItem.Create(nil);
  fMsg:= TIdMessage.Create(Self);
  fSMTP:= TIdSMTP.Create(Self);
  fUseTLS:=utNoTLSSupport;
  with fSMTP do begin
    ConnectTimeout:=defSmtpTimeOut;
    ReadTimeout:=defSmtpTimeOut;
    fCertPwd:='';
  // if openSSL libraries and certificates are available, use SSL for TLS support
    fTLSAvail:=(length(Root)>0) and (length(Cert)>0) and (length(Key)>0);
    if fTLSAvail then begin
      try
        TIdSSLContext.Create.Free;
        fTLSAvail:=true;
      except
        fTLSAvail:=false;
        end;
      if fTLSAvail then begin
        sr:=Path+Root; sc:=Path+Cert; sk:=Path+Key;
        fTLSAvail:=FileExists(sr) and FileExists(sc) and FileExists(sk);
        end;
      end;
    if fTLSAvail then begin
      IOHandler:=TIdSSLIOHandlerSocketOpenSSL.Create(fSmtp);
      fCertPwd:=Pwd;
      with (fSMTP.IOHandler as TIdSSLIOHandlerSocketOpenSSL)do begin
        with SSLOptions do begin
          RootCertFile:=sr; CertFile:=sc; KeyFile:=sk;
          end;
        OnGetPassword:=GetCertPassword;
        end;
      end
    else IOHandler:=TIdIOHandler.MakeDefaultIOHandler(fSmtp);
    ManagedIOHandler:=True;
    UseTLS:=utNoTLSSupport;
    OnConnected:=IsConnected;
    OnDisconnected:=Disconnected;
    end;
  fHTML:=false;
  fPriority:= mpNormal;
  fExtraHeaders:=TIdHeaderList.Create;
  fAbout:= 'Version 3.4, 2005-2008  Mats Asplund, J. Rathlev';
  fOnSendError:=nil;
  fOnConnected:=nil;
  fOnDisconnected:=nil;
  fGetSSLPassword:=nil;
  end;

constructor TmMailer.Create (AOwner : TComponent);
begin
  Create(AOwner,'','','','','');
  end;

destructor TmMailer.Destroy;
begin
  fAttachments.fAttachedFiles.Free; fAttachments.Free;
  fBody.Free; fHTMLBody.Free; fMsg.Free;
  fRecipients.Free; fBCC.Free; fFrom.Free;
  fExtraHeaders.Free;
  fSMTP.IOHandler.Free;
  fSMTP.Free;
  inherited Destroy;
  end;

procedure TmMailer.GetCertPassword(var Password : string);
begin
  if Assigned(fGetSSLPassword) then fGetSSLPassword(Password)  // prompt for password
  else Password:=fCertPwd;
  end;

procedure TmMailer.IsConnected(Sender: TObject);
begin
  if Assigned(fOnConnected) then fOnConnected(self);
end;

procedure TmMailer.Disconnected(Sender: TObject);
begin
  if Assigned(fOnDisconnected) then fOnDisconnected(self);
end;

procedure TmMailer.SetTLS(Value : integer);
begin
  if fTLSAvail and (Value>0) then begin
    case Value of
    2 : fUseTLS:=utUseRequireTLS;
    3 : fUseTLS:=utUseImplicitTLS;
    else fUseTLS:=utUseExplicitTLS;
      end;
    end
  else fUseTLS:=utNoTLSSupport;
  end;

procedure TmMailer.ClearExtraHeaders;
begin
  fExtraHeaders.Clear;
  end;

procedure TmMailer.AddExtraHeader (fId,fContents : string);
begin
  fExtraHeaders.Values[fId]:=fContents;
  end;

function TmMailer.Send : boolean;
var
  n,k : integer;
  s  : string;
  at : TIdAttachmentFile;

  procedure Authenticate;
  begin
    if not fSMTP.Authenticate then
      raise EIdException.Create(RSSMTPSvrAuthFailed);
    end;

begin
  Result:=true;
  fMsg.ClearBody;               // inserted, J. Rathlev, Jan. 2004
  // handling of multipart messages, J. Rathlev, Jun.. 2007
  if fHTML and (fBody.Count>0) then begin
  // multipart/alternative if HTML, text and Attachemets
    with TIdText.Create(fMsg.MessageParts,fBody) do begin
      ContentType:='multipart/alternative';  // Container for text parts
      end;
    with TIdText.Create(fMsg.MessageParts,fBody) do begin
      ContentType:=defTextType;
      ContentTransfer:=defEncoding;
      ParentPart:=0;
      end;
    with TIdText.Create(fMsg.MessageParts,fHTMLBody) do begin
      ContentType:=defHtmlType;
      ContentTransfer:=defEncoding;
      ParentPart:=0;
      end;
    end
  else if fHTML then begin
    with TIdText.Create(fMsg.MessageParts,fHTMLBody) do begin
      ContentType:=defHtmlType;
      ContentTransfer:=defEncoding;
      end;
    end
  else begin
    with TIdText.Create(fMsg.MessageParts,fBody) do begin
      ContentType:=defTextType;
      ContentTransfer:=defEncoding;
      end;
    end;
  with fAttachments do if AttachedFiles.Count>0 then begin // message with attachements
    for n:= 0 to AttachedFiles.Count -1 do begin  // attachments
      s:=AttachedFiles[n];
      k:=pos('|',s);                       // separate filename and content type
      if k=0 then k:=length(s)+1;
      at:=TIdAttachmentFile.Create(fMsg.MessageParts,copy(s,1,k-1));
      delete(s,1,k);
      if length(s)>0 then begin
        k:=pos('/',s);   // other than application?
        if k=0 then at.ContentType:='application/'+s
        else at.ContentType:=s;
        end;
      end;
    end;

  fMsg.Recipients:= fRecipients;
  fMsg.BccList:= fBCC;
  fMsg.From:= fFrom;
  fMsg.Subject:= fSubject;
  fMsg.ExtraHeaders:=fExtraHeaders;

  if (length(fUserId)>0) and (length(fPassword)>0) then fSMTP.AuthType:=atDefault
  else fSMTP.AuthType:=atNone;
  fSMTP.UserName:= fUserId;
  fSMTP.Password:= fPassword;
  fSMTP.UseTLS:=fUseTLS;

  fSMTP.Host:= fHost;
  fSMTP.Port:= fPort;
  try
    fSMTP.Connect;
    Authenticate;
    try
      fSMTP.Send(fMsg);
    finally
      fSMTP.Disconnect;
      end;
  except
    on E: Exception do begin
      if Assigned(fOnSendError) then begin
        s:=E.Message;
        s:=s+'<'+fRecipients.EMailAddresses+'>';
        fOnSendError(Self,s);
        end;
      Result:=false;
      fSMTP.Disconnect;
      end;
    end;
  end;

procedure TmMailer.SetBody(const Value: TStrings);
begin
  fBody.Assign(Value);
end;

procedure TmMailer.SetHTMLBody(const Value: TStrings);
begin
  fHTMLBody.Assign(Value);
end;

procedure Register;
begin
  RegisterComponents('Internet', [TmMailer]);
end;

end.

